@@ -27,7 +27,7 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * $Id: chalib.c,v 1.46 2003/08/15 01:56:51 kazuma-t Exp $
+ * $Id: chalib.c,v 1.3 2007/03/25 11:57:27 kazuma-t Exp $
  */
 
 #include "chalib.h"
@@ -37,6 +37,9 @@
 
 #define CHA_NAME       "ChaSen"
 
+#define STR_UNSPECIFIED "UNSPEC"
+#define STR_ANNOTATION "ANNO"
+
 darts_t *Da_dicfile[DIC_NUM];
 int Da_ndicfile = 0;
 
@@ -50,7 +53,7 @@ enum cha_encode Cha_encode = CHASEN_ENCODE_EUCJP;
 
 static int cost_width0;
 
-static char dadic_filename[DIC_NUM][CHA_FILENAME_LEN];
+static char dadic_filename[DIC_NUM][PATH_MAX];
 
 static int opt_show = 'b', opt_form = 'f', opt_ja, opt_cmd;
 static char *opt_form_string;
@@ -65,7 +68,7 @@ cha_version(FILE * fp)
 	return;
 
     fprintf(fp,
-	    "%s version %s (c) 1996-2001 Nara Institute of Science and Technology\n",
+	    "%s version %s (c) 1996-2007 Nara Institute of Science and Technology\n",
 	    CHA_NAME, VERSION);
     fprintf(fp, "Grammar files are in ChaSen's new v-gram format.\n");
 }
@@ -278,10 +281,11 @@ chasen_getopt_argv(char **argv, FILE * fp)
  *     1 - quit
  */
 static int
-chasen_sparse_main(char *input, FILE * output)
+chasen_sparse_main(char *input, FILE *output)
 {
     char *crlf;
-
+    cha_lat_t lat;
+    cha_seg_t seg;
     /*
      * initialize if not done 
      */
@@ -297,18 +301,11 @@ chasen_sparse_main(char *input, FILE * output)
 	return 0;
     }
 
-    /*
-     * conversion of ISO-2022-JP string to EUC-JP 
-     */
-    /*
-     * jis_to_euc(input);
-     */
-
     /*
      * parse a sentence and print 
      */
     while (*input) {
-	int c = 0, len;
+	int c = 0, len, cursor;
 	if ((crlf = strpbrk(input, "\r\n")) == NULL)
 	    len = strlen(input);
 	else {
@@ -317,10 +314,23 @@ chasen_sparse_main(char *input, FILE * output)
 	    *crlf = '\0';
 	}
 	cha_print_reset();
-	if (len > 0 && !cha_parse_sentence(input, len)) {
-	    cha_print_path(opt_show, opt_form, opt_form_string);
-	} else if (!opt_ja)
-	    cha_print_bos_eos(opt_form);
+
+	cursor = 0;
+	cha_parse_bos(&lat);
+	while (cursor < len) {
+	    seg.text = input + cursor;
+	    seg.anno_no = -1;
+	    seg.len = cha_tok_parse(Cha_tokenizer, seg.text, seg.char_type,
+				    len - cursor, &seg.anno_no);
+	    if (seg.anno_no >= 0)
+		seg.type = SEGTYPE_ANNOTATION;
+	    else
+		seg.type = SEGTYPE_NORMAL;
+	    cha_parse_segment(&lat, &seg);
+	    cursor += seg.len;
+	}
+	cha_parse_eos(&lat);
+	cha_print_path(&lat, opt_show, opt_form, opt_form_string);
 
 	if (crlf == NULL)
 	    break;
@@ -330,7 +340,6 @@ chasen_sparse_main(char *input, FILE * output)
 	    input = crlf + 1;
     }
 
-
     return 0;
 }
 
@@ -371,6 +380,145 @@ chasen_sparse(char *str_in, FILE * fp_out)
     return rc;
 }
 
+static int
+set_normal(cha_seg_t *seg)
+{
+    seg->type = SEGTYPE_NORMAL;
+    cha_tok_parse(Cha_tokenizer, seg->text, seg->char_type,
+		  seg->len, NULL);
+
+    return seg->len;
+}
+
+static int
+seg_tokenize(unsigned char *line, cha_seg_t *seg)
+{
+    int len;
+    int i;
+
+    len = 0;
+    seg->text = line;
+    while (line[len] != '\t' && line[len] != '\0')
+	len++;
+    seg->len = len;
+    seg->posid = seg->inf_type = seg->inf_form = 0;
+
+    if (line[len] == '\0')
+	return set_normal(seg);
+
+    /* skip reading and base form */
+    for (i = 0; i < 2; i++) {
+	len++;
+	while (line[len] != '\t' && line[len] != '\0')
+	    len++;
+	if (line[len] == '\0')
+	    return set_normal(seg);
+    }
+
+    line += len + 1;
+    if (strcmp(line, STR_UNSPECIFIED) == 0) {
+	seg->type = SEGTYPE_UNSPECIFIED;
+	seg->char_type[0] = seg->len;
+    } else if (strcmp(line, STR_ANNOTATION) == 0) {
+	seg->type = SEGTYPE_ANNOTATION;
+	cha_tok_parse(Cha_tokenizer, seg->text, seg->char_type,
+		      seg->len, &seg->anno_no);
+	seg->char_type[0] = seg->len;
+    } else { /* read POS */
+	char *pos[256], *itype;
+	char *l = line;
+
+	seg->type = SEGTYPE_MORPH;
+	seg->char_type[0] = seg->len;
+	if ((l = strchr(l, '\t')) != NULL) {
+	    *l++ = '\0';
+	    itype = l;
+	    if ((l = strchr(l, '\t')) != NULL) {
+		*l++ = '\0';
+		seg->inf_type = cha_get_type_id(itype);
+		seg->inf_form = cha_get_form_id(l, seg->inf_type);
+	    } else {
+		fprintf(stderr, "invalid format: %s\n", line);
+		return -1;
+	    }
+	}
+	i = 0;
+	pos[i++] = l = line;
+	while ((l = strchr(l, '-')) != NULL) {
+	    *l++ = '\0';
+	    pos[i++] = l;
+	}
+	pos[i] = NULL;
+	if (cha_litmatch(pos[0], 3, STR_UNKNOWN_WORD,
+			 STR_UNKNOWN_WORD1, STR_UNKNOWN_WORD2))
+	    seg->is_undef = 1;
+	else {
+	    seg->is_undef = 0;
+	    seg->posid = cha_get_nhinsi_str_id(pos);
+	}
+    }
+    
+    return seg->len;
+}
+
+static int
+strip(unsigned char *s)
+{
+    int len = strlen(s);
+
+    if (s[len - 1] == '\n')
+	s[len-- - 1] = '\0';
+
+    while (len > 0 && s[len - 1] == '\t')
+        s[len-- - 1] = '\0';
+
+    return len;
+}
+
+
+int
+chasen_parse_segments(FILE *input, FILE *output)
+{
+    cha_lat_t lat;
+    unsigned char buf[CHA_INPUT_SIZE]; /* XXX */
+    cha_seg_t seg;
+    int is_eos = 1;
+
+    if (!Cha_undef_info_num)
+	cha_init();
+    if (!opt_form_string)
+	cha_set_opt_form(NULL);
+
+    cha_set_output(output);
+
+    while (fgets(buf, CHA_INPUT_SIZE, input) != NULL) {
+	strip(buf);
+	if (is_eos) {
+	    cha_print_reset();
+	    cha_parse_bos(&lat);
+	    is_eos = 0;
+	}
+	if (!buf[0] || cha_litmatch(buf, 2, STR_EOS, STR_BOS_EOS)) {
+	    /* EOS */
+	    cha_parse_eos(&lat);
+	    cha_print_path(&lat, opt_show, opt_form, opt_form_string);
+	    is_eos = 1;
+	    continue;
+	}
+	if (seg_tokenize(buf, &seg) < 0) {
+	    fprintf(stderr, "invalid format: %s\n", buf);
+	    continue;
+	}
+	cha_parse_segment(&lat, &seg);
+    }
+    if (!is_eos) {
+        cha_parse_eos(&lat);
+	cha_print_path(&lat, opt_show, opt_form, opt_form_string);
+    }
+
+    return lat.len;
+}
+
 /*
  * read from file/str, parse, and output to string
  * 
@@ -425,26 +573,26 @@ cha_fgets(char *s, int n, FILE * fp)
 }
 
 static void
-set_dic_filename(char *filename, char *s)
+set_dic_filename(char *filename, size_t len, char *s)
 {
 #ifdef PATHTYPE_MSDOS
     if (*s == PATH_DELIMITER || *s && s[1] == ':')
-	strcpy(filename, s);
+	strncpy(filename, s, len);
 #else
     if (*s == PATH_DELIMITER)
-	strcpy(filename, s);
+	strncpy(filename, s, len);
 #endif /* PATHTYPE_MSDOS */
     else
-	sprintf(filename, "%s%s", cha_get_grammar_dir(), s);
+	snprintf(filename, len, "%s%s", cha_get_grammar_dir(), s);
 }
 
 void
 cha_read_dadic(chasen_cell_t * cell)
 {
     int num;
-    char da_filename[CHA_FILENAME_LEN];
-    char lex_filename[CHA_FILENAME_LEN];
-    char dat_filename[CHA_FILENAME_LEN];
+    char da_filename[PATH_MAX];
+    char lex_filename[PATH_MAX];
+    char dat_filename[PATH_MAX];
 
     if (dadic_filename[0][0])
 	return;
@@ -452,11 +600,12 @@ cha_read_dadic(chasen_cell_t * cell)
     for (num = 0; !nullp(cell); num++, cell = cha_cdr(cell)) {
 	if (num >= DIC_NUM)
 	    cha_exit_file(1, "too many Darts dictionary files");
-	set_dic_filename(dadic_filename[num], cha_s_atom(cha_car(cell)));
+	set_dic_filename(dadic_filename[num], PATH_MAX,
+			 cha_s_atom(cha_car(cell)));
 
-	sprintf(da_filename, "%s.da", dadic_filename[num]);
-	sprintf(lex_filename, "%s.lex", dadic_filename[num]);
-	sprintf(dat_filename, "%s.dat", dadic_filename[num]);
+	snprintf(da_filename, PATH_MAX, "%s.da", dadic_filename[num]);
+	snprintf(lex_filename, PATH_MAX, "%s.lex", dadic_filename[num]);
+	snprintf(dat_filename, PATH_MAX, "%s.dat", dadic_filename[num]);
 	Da_dicfile[num] = da_open(da_filename,
 				  lex_filename, dat_filename);
     }
