@@ -27,7 +27,7 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * $Id: print.c,v 1.53 2003/08/11 05:53:36 kazuma-t Exp $
+ * $Id: print.c,v 1.3 2007/03/25 11:02:53 kazuma-t Exp $
  */
 
 #include <stdarg.h>
@@ -47,7 +47,7 @@
 static int path_buffer[CHA_INPUT_SIZE];
 static int pos_end = 0;
 
-static void (*cha_putc) (), (*cha_puts) (), (*cha_printf) ();
+static void (*cha_putc) (), (*cha_puts) (), (*cha_printf) (), (*cha_write) ();
 
 void
 cha_print_reset(void)
@@ -59,14 +59,14 @@ cha_print_reset(void)
 static void
 dump_mrph_data(mrph_data_t *data)
 {
-    fprintf(stdout, "hw:\t`%s`\n", data->headword);
-    fprintf(stdout, "hw_len:\t%d\n", data->headword_len);
-    fprintf(stdout, "undef:\t%d\n", data->is_undef);
-    fprintf(stdout, "pos:\t%d\n", data->posid);
-    fprintf(stdout, "itype:\t%d\n", data->inf_type);
-    fprintf(stdout, "iform:\t%d\n", data->inf_form);
-    fprintf(stdout, "contbl:\t%d\n", data->con_tbl);
-    fprintf(stdout, "weight:\t%d\n", data->weight);
+    fprintf(stdout, "hw:\t`%s`\n", data->mrph->headword);
+    fprintf(stdout, "hw_len:\t%d\n", data->mrph->headword_len);
+    fprintf(stdout, "undef:\t%d\n", data->mrph->is_undef);
+    fprintf(stdout, "pos:\t%d\n", data->mrph->posid);
+    fprintf(stdout, "itype:\t%d\n", data->mrph->inf_type);
+    fprintf(stdout, "iform:\t%d\n", data->mrph->inf_form);
+    fprintf(stdout, "contbl:\t%d\n", data->mrph->con_tbl);
+    fprintf(stdout, "weight:\t%d\n", data->mrph->weight);
     fprintf(stdout, "stem_len:\t%d\n", data->stem_len);
     fprintf(stdout, "reading:\t`%s`\n", data->reading);
     fprintf(stdout, "pron:\t`%s`\n", data->pron);
@@ -80,13 +80,13 @@ dump_mrph_data(mrph_data_t *data)
 static mrph_data_t *
 get_mrph_data(mrph_t *mrph, mrph_data_t *data)
 {
-    memcpy(data, mrph, sizeof(mrph_t));
+    data->mrph = mrph;
 
-    if (mrph->dat_index < 0) {
+    if (mrph->darts == NULL) {
 	data->reading = data->pron = data->base
 	    = data->info = "";
+	data->stem_len = data->reading_len = data->pron_len = -1;
 	data->compound = NO_COMPOUND;
-	data->stem_len = 0;
     } else {
 	da_dat_t d;
 	void *p;
@@ -166,6 +166,12 @@ cha_sprintf(char *output, char *format, ...)
     cha_sputs(tmpbuf, output);
 }
 
+static void
+cha_swrite(void *ptr, size_t size, size_t nmemb, char *output)
+{
+    memcpy(output, ptr, size * nmemb);
+}
+
 void
 cha_set_output(FILE * output)
 {
@@ -186,6 +192,7 @@ cha_set_output(FILE * output)
 	cha_putc = (void (*)) cha_sputc;
 	cha_puts = (void (*)) cha_sputs;
 	cha_printf = (void (*)) cha_sprintf;
+	cha_write = (void (*)) cha_swrite;
 	/*
 	 * initialize output buffer 
 	 */
@@ -205,6 +212,7 @@ cha_set_output(FILE * output)
 	cha_putc = (void (*)) fputc;
 	cha_puts = (void (*)) fputs;
 	cha_printf = (void (*)) fprintf;
+	cha_write = (void (*)) fwrite;
     }
 }
 
@@ -326,6 +334,11 @@ set_ruby(char *dest, char *midasi, char *yomi,
     int stat = 0;
     int plen, maxplen = 0;
 
+    if (yomi[0] == '\0') {
+	strcpy(d, m);
+	return;
+    }
+
     for (;;) {
 	for (; *y; y += cha_tok_mblen(Cha_tokenizer, y, 4 /* XXX */)) {
 	    if (stat == 0) {
@@ -524,15 +537,14 @@ get_deli_right(int c)
 }
 
 static void
-print_anno(int path_num, char *format)
+print_anno(cha_lat_t *lat, int path_num, char *format)
 {
     path_t *path = &Cha_path[path_num];
-    mrph_data_t mrph;
+    mrph_t mrph;
+    mrph_data_t data;
     int start, end;
 
-    if (!Cha_anno_info[0].hinsi && !Cha_anno_info[1].hinsi
-	&& !Cha_anno_info[1].format)
-	return;
+    data.mrph = &mrph;
 
     if (path->start <= pos_end) {
 	pos_end = path->end;
@@ -542,82 +554,65 @@ print_anno(int path_num, char *format)
     start = path->start;
     end = path->end;
 
-    while (start > pos_end) {
-	int anno_no = cha_tok_anno_type(Cha_tokenizer, pos_end);
-	char *format_string = format;
-	if (anno_no >= 0 &&
-	    (Cha_anno_info[anno_no].hinsi
-	     || Cha_anno_info[anno_no].format)) {
-	    mrph.headword = Cha_tokenizer->string + pos_end;
-	    mrph.headword_len = mrph.stem_len =
-		cha_tok_char_type_len(Cha_tokenizer, pos_end);
-	    mrph.reading_len = -1;
-	    mrph.pron_len = -1;
-	    mrph.reading = "";
-	    mrph.pron = "";
-	    mrph.base = "";
-	    if (Cha_anno_info[anno_no].format) {
-		format_string = Cha_anno_info[anno_no].format;
-		mrph.is_undef = 1;
-		mrph.posid = Cha_undef_info[0].hinsi;
-	    } else {
-		mrph.is_undef = 0;
-		mrph.posid = Cha_anno_info[anno_no].hinsi;
-	    }
-	    mrph.con_tbl = 0;
-	    mrph.inf_type = 0;
-	    mrph.inf_form = 0;
-	    mrph.weight = 0;
-	    mrph.info = "";
-	    path->start = pos_end;
-	    path->end = pos_end + mrph.headword_len;
-	    cha_printf_mrph(path_num, &mrph, format_string);
-	}
-	pos_end += cha_tok_char_type_len(Cha_tokenizer, pos_end);
-    }
-    path->end = pos_end = end;
-    path->start = start;
-}
+    if (lat->anno >= 0) {
+	while (start > pos_end && lat->anno >= 0) {
+	    mrph_t *m = nth_mrph(lat->anno);
+	    mrph_data_t d;
+	    int anno_no = m->con_tbl; /* XXX */
 
-#define copy_mbstr(d,s) \
-do { int len = cha_tok_mblen(Cha_tokenizer, (s), 4 /* XXX */); \
-    while (len-- && *(s)) { *(d)++ = *(s)++; } \
-    s--; /* XXX */\
-} while (0)
+	    path->start = pos_end;
+	    path->end = pos_end + m->headword_len;
+	    get_mrph_data(m, &d);
+
+	    if (Cha_anno_info[anno_no].format)
+		cha_printf_mrph(lat, path_num, &d,
+				Cha_anno_info[anno_no].format);
+	    else if (Cha_anno_info[anno_no].hinsi)
+		cha_printf_mrph(lat, path_num, &d, format);
+
+	    lat->anno = m->dat_index; /* XXX */
+	    pos_end += m->headword_len;
+	    continue;  
+	}
+	path->end = pos_end = end;
+	path->start = start;
+    }
+}
 
 static void
 extract_yomi1(char *dst, char *src)
 {
-    int in_brace = 0, is1st = 0;
-    char *s, *d;
+    char *d = dst;
+    char *s = src;
+    int state = 0;
 
-    if (strchr(src, '{') == NULL) {
-	if (dst != src)
-	    strcpy(dst, src);
-	return;
-    }
+    while (*s) {
+	int i;
+	int len = cha_tok_mblen(Cha_tokenizer, (s), 4 /* XXX */);
 
-    for (s = src, d = dst; *s; s++) {
-	if (!in_brace) {
-	    if (*s == '{')
-		in_brace = is1st = 1;
-	    else
-		copy_mbstr(d, s);
-	} else if (*s == '}')
-	    in_brace = 0;
-	else if (is1st) {
-	    if (*s == '/')
-		is1st = 0;
-	    else
-		copy_mbstr(d, s);
+	if (state == 0 && len == 1 && *s == '{') {
+	    state = 1;
+	    s += len;
+	    continue;
+	} else if (state == 1 && len == 1 && *s == '/') {
+	    state = 2;
+	    s += len;
+	    continue;
+	} else if (state == 2) {
+	    if (len == 1 && *s == '}')
+		state = 0;
+	    s += len;
+	    continue;
 	}
-    }
 
+	for (i = 0; i < len; i++)
+	    *d++ = *s++;
+    }
     *d = '\0';
 }
 
 void
-cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
+cha_printf_mrph(cha_lat_t *lat, int path_num, mrph_data_t *mdata, char *format)
 {
     int letter, value, n, state;
     int deli_left = 0, deli_right = 0;
@@ -625,6 +620,7 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
     char word_str[CHA_INPUT_SIZE], word_str2[CHA_INPUT_SIZE];
     char *width_str;
     path_t *path = &Cha_path[path_num];
+    mrph_t *mrph = mdata->mrph;
 
     eword = NULL;		/* string in EUC */
     word = NULL;		/* string in EUC(UNIX) or SJIS(Win) */
@@ -675,7 +671,7 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 		    state = 1;
 		break;
 	    case 'I':
-		if (mrph->info[0] && strcmp(mrph->info, "NIL"))
+		if (mdata->info[0] && strcmp(mdata->info, "NIL"))
 		    state = 1;
 		break;
 	    case 'T':
@@ -714,14 +710,16 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 	    word_str[mrph->headword_len] = '\0';
 	    break;
 	case 'M':		/* Surface string (base form) */
-	    if (mrph->base[0])
-		eword = mrph->base;
+	    if (mdata->base[0])
+		eword = mdata->base;
+	    else if (mdata->stem_len < 0)
+		eword = "";
 	    else {
-		memcpy(eword = word_str, mrph->headword, mrph->stem_len);
+		memcpy(eword = word_str, mrph->headword, mdata->stem_len);
 		if (!mrph->inf_type)
-		    word_str[mrph->stem_len] = '\0';
+		    word_str[mdata->stem_len] = '\0';
 		else
-		    strcpy(word_str + mrph->stem_len,
+		    strcpy(word_str + mdata->stem_len,
 			   Cha_form[mrph->inf_type][Cha_type[mrph->inf_type].
 						    basic].gobi);
 	    }
@@ -730,12 +728,12 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 	case 'Y':		/* Japanese Reading (base form) */
 	case 'r':
 	    word_str[0] = '\0';
-	    if (mrph->reading_len >= 0) {
+	    if (mdata->reading_len >= 0) {
 		if (s[0] != 'r' && s[1] != '0')
-		    extract_yomi1(word_str, mrph->reading);
+		    extract_yomi1(word_str, mdata->reading);
 		else
-		    strcpy(word_str, mrph->reading);
-		if (mrph->inf_type > 0 && mrph->stem_len >= 0) {
+		    strcpy(word_str, mdata->reading);
+		if (mrph->inf_type > 0 && mdata->stem_len >= 0) {
 		    if (*s != 'Y')
 			strcat(word_str,
 			       Cha_form[mrph->inf_type]
@@ -771,12 +769,12 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 	case 'a':		/* Japanese pronunciation (surface form) */
 	case 'A':		/* Japanese pronunciation (base form) */
 	    word_str[0] = '\0';
-	    if (mrph->pron_len >= 0) {
+	    if (mdata->pron_len >= 0) {
 		if (s[1] != '0')
-		    extract_yomi1(word_str, mrph->pron);
+		    extract_yomi1(word_str, mdata->pron);
 		else
-		    strcpy(word_str, mrph->pron);
-		if (mrph->inf_type > 0 && mrph->stem_len >= 0) {
+		    strcpy(word_str, mdata->pron);
+		if (mrph->inf_type > 0 && mdata->stem_len >= 0) {
 		    if (*s != 'A')
 			strcat(word_str, Cha_form[mrph->inf_type]
 			       [mrph->inf_form].pgobi);
@@ -791,16 +789,18 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 	    break;
 	case 'i':		/* information */
 	    if (s[1] != '0')
-		extract_yomi1(word_str, mrph->info);
+		extract_yomi1(word_str, mdata->info);
 	    else
-		strcpy(word_str, mrph->info);
+		strcpy(word_str, mdata->info);
 	    eword = word_str;
+	    if (s[1] == '0' || s[1] == '1')
+		s++;
 	    break;
 	case 'I':		/* information */
 	    if (*++s == '\0')
 		cha_putc(*--s, cha_output);
-	    else if (mrph->info[0] && strcmp(mrph->info, "NIL"))
-		eword = mrph->info;
+	    else if (mdata->info[0] && strcmp(mdata->info, "NIL"))
+		eword = mdata->info;
 	    else if (*s != ' ')
 		letter = *s;
 	    break;
@@ -879,7 +879,8 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 	    value *= mrph->weight * Cha_mrph_cost_weight;
 	    break;
 	case 'S':		/* entire sentence */
-	    word = Cha_tokenizer->string;
+	    memcpy(eword = word_str, lat->text, lat->len);
+	    word_str[lat->len] = '\0';
 	    break;
 	case 'p':		/* the information about path */
 	    if (s[1] == '\0') {
@@ -965,32 +966,32 @@ cha_printf_mrph(int path_num, mrph_data_t *mrph, char *format)
 }
 
 static void
-print_bos_eos(char *str)
+print_bos_eos(cha_lat_t *lat, char *str)
 {
     char *s;
 
     for (s = str; *s; s++) {
-	if (*s == '%' && *++s == 'S')
-	    cha_puts(Cha_tokenizer->string, cha_output);
+	if (*s == '%' && *++s == 'S' && lat)
+	    cha_write(lat->text, sizeof(char), lat->len, cha_output);
 	else
 	    cha_putc(*s, cha_output);
     }
 }
 
 static void
-print_bos(int opt_form)
+print_bos(cha_lat_t *lat, int opt_form)
 {
     if (opt_form != 'W' && opt_form != 'd' && *Cha_bos_string)
-	print_bos_eos(Cha_bos_string);
+	print_bos_eos(lat, Cha_bos_string);
 }
 
 static void
-print_eos(int opt_form)
+print_eos(cha_lat_t *lat, int opt_form)
 {
     if (opt_form == 'W')
 	cha_putc('\n', cha_output);
     else if (opt_form != 'd' && *Cha_eos_string)
-	print_bos_eos(Cha_eos_string);
+	print_bos_eos(lat, Cha_eos_string);
 }
 
 static int
@@ -1017,42 +1018,44 @@ get_compound(mrph_data_t *data, char *headword, darts_t *da, long index)
  * print_path_mrph
  */
 static void
-print_mrph(int path_num, mrph_data_t *mrph, char *format)
+print_mrph(cha_lat_t *lat, int path_num, mrph_data_t *mdata, char *format)
 {
-    print_anno(path_num, format);
+    mrph_t *mrph = mdata->mrph;
 
-    if (Cha_output_iscompound || mrph->compound == NO_COMPOUND) {
-        cha_printf_mrph(path_num, mrph, format); 
+    print_anno(lat, path_num, format);
+
+    if (Cha_output_iscompound || mdata->compound == NO_COMPOUND) {
+        cha_printf_mrph(lat, path_num, mdata, format); 
     } else {
 	mrph_data_t data;
-	long index = mrph->compound;
+	long index = mdata->compound;
 	int has_next = 1;
 	char *headword = mrph->headword;
 
 	while (has_next) {
 	    has_next = get_compound(&data, headword, mrph->darts, index);
 	    if (!has_next) {
-		if (mrph->inf_type > 0 && mrph->stem_len >= 0) {
-		    data.inf_form = mrph->inf_form;
-		    data.headword_len +=
+		if (mrph->inf_type > 0 && mdata->stem_len >= 0) {
+		    data.mrph->inf_form = mrph->inf_form;
+		    data.mrph->headword_len +=
 			strlen(Cha_form[mrph->inf_type][mrph->inf_form].gobi);
 		}
-		data.con_tbl = mrph->con_tbl;
+		data.mrph->con_tbl = mrph->con_tbl;
 	    }
-	    cha_printf_mrph(path_num, &data, format);
-	    headword += data.headword_len;
+	    cha_printf_mrph(lat, path_num, &data, format);
+	    headword += data.mrph->headword_len;
 	    index += sizeof(short) * 2 + sizeof(da_lex_t);
 	}
     }
 }
 
 static void
-print_path_mrph(int path_num, char *format)
+print_path_mrph(cha_lat_t *lat, int path_num, char *format)
 {
     mrph_data_t data;
 
     get_mrph_data(nth_mrph(Cha_path[path_num].mrph_p), &data);
-    print_mrph(path_num, &data, format);
+    print_mrph(lat, path_num, &data, format);
 }
 
 static void
@@ -1061,10 +1064,11 @@ concat_composit_mrph(mrph_data_t *composit_mrph, mrph_data_t *cur_mrph)
     /* 
      * initialization
      */
-    if (!composit_mrph->posid) {    
-	composit_mrph->posid = Cha_hinsi[cur_mrph->posid].composit;
-	composit_mrph->headword = cur_mrph->headword;
-	composit_mrph->headword_len = composit_mrph->weight = 0;
+    if (!composit_mrph->mrph->posid) {    
+	composit_mrph->mrph->posid
+	    = Cha_hinsi[cur_mrph->mrph->posid].composit;
+	composit_mrph->mrph->headword = cur_mrph->mrph->headword;
+	composit_mrph->mrph->headword_len = composit_mrph->mrph->weight = 0;
 	composit_mrph->reading[0] = '\0';
 	composit_mrph->pron[0] = '\0';
 	composit_mrph->base[0] = '\0';
@@ -1077,12 +1081,13 @@ concat_composit_mrph(mrph_data_t *composit_mrph, mrph_data_t *cur_mrph)
     else {
 	int len = strlen(composit_mrph->reading);
 	memcpy(composit_mrph->reading + len,
-	       cur_mrph->headword, cur_mrph->stem_len);
+	       cur_mrph->mrph->headword, cur_mrph->stem_len);
 	composit_mrph->reading[len + cur_mrph->stem_len] = '\0';
     }
-    if (cur_mrph->inf_type > 0)
+    if (cur_mrph->mrph->inf_type > 0)
 	strcat(composit_mrph->reading,
-	       Cha_form[cur_mrph->inf_type][cur_mrph->inf_form].ygobi);
+	       Cha_form
+	       [cur_mrph->mrph->inf_type][cur_mrph->mrph->inf_form].ygobi);
     /* 
      * Pronunciation
      */
@@ -1093,16 +1098,17 @@ concat_composit_mrph(mrph_data_t *composit_mrph, mrph_data_t *cur_mrph)
     else {
 	int len = strlen(composit_mrph->pron);
 	memcpy(composit_mrph->pron + len,
-	       cur_mrph->headword, cur_mrph->stem_len);
+	       cur_mrph->mrph->headword, cur_mrph->stem_len);
 	composit_mrph->pron[len + cur_mrph->stem_len] = '\0';
     }
-    if (cur_mrph->inf_type > 0)
+    if (cur_mrph->mrph->inf_type > 0)
 	strcat(composit_mrph->pron,
-	       Cha_form[cur_mrph->inf_type][cur_mrph->inf_form].pgobi);
+	       Cha_form
+	       [cur_mrph->mrph->inf_type][cur_mrph->mrph->inf_form].pgobi);
 
     strcat(composit_mrph->base, cur_mrph->base);
-    composit_mrph->headword_len += cur_mrph->headword_len;
-    composit_mrph->weight += cur_mrph->weight;
+    composit_mrph->mrph->headword_len += cur_mrph->mrph->headword_len;
+    composit_mrph->mrph->weight += cur_mrph->mrph->weight;
 }
 
 static void
@@ -1115,7 +1121,7 @@ concat_composit_mrph_end(mrph_data_t *composit_mrph, mrph_data_t *cur_mrph)
         strcat(composit_mrph->reading, cur_mrph->reading);
     else {
         int len = strlen(composit_mrph->reading);
-        memcpy(composit_mrph->reading + len, cur_mrph->headword,
+        memcpy(composit_mrph->reading + len, cur_mrph->mrph->headword,
     	   cur_mrph->stem_len);
         composit_mrph->reading[len + cur_mrph->stem_len] = '\0';
     }
@@ -1128,29 +1134,29 @@ concat_composit_mrph_end(mrph_data_t *composit_mrph, mrph_data_t *cur_mrph)
         strcat(composit_mrph->pron, cur_mrph->reading);
     else {
         int len = strlen(composit_mrph->pron);
-        memcpy(composit_mrph->pron + len, cur_mrph->headword,
+        memcpy(composit_mrph->pron + len, cur_mrph->mrph->headword,
     	   cur_mrph->stem_len);
         composit_mrph->pron[len + cur_mrph->stem_len] = '\0';
     }
 
     strcat(composit_mrph->base, cur_mrph->base);
     composit_mrph->stem_len =
-	composit_mrph->headword_len + cur_mrph->stem_len;
-    composit_mrph->headword_len += cur_mrph->headword_len;
-    composit_mrph->weight += cur_mrph->weight;
+	composit_mrph->mrph->headword_len + cur_mrph->stem_len;
+    composit_mrph->mrph->headword_len += cur_mrph->mrph->headword_len;
+    composit_mrph->mrph->weight += cur_mrph->mrph->weight;
     composit_mrph->info = cur_mrph->info;
-    composit_mrph->inf_type = cur_mrph->inf_type;
-    composit_mrph->inf_form = cur_mrph->inf_form;
-    composit_mrph->is_undef = cur_mrph->is_undef;
+    composit_mrph->mrph->inf_type = cur_mrph->mrph->inf_type;
+    composit_mrph->mrph->inf_form = cur_mrph->mrph->inf_form;
+    composit_mrph->mrph->is_undef = cur_mrph->mrph->is_undef;
 }
 
 #define print_anno_eos() \
-    { print_anno(Cha_path_num - 1, format); print_eos(opt_form); }
+    { print_anno(lat, Cha_path_num - 1, format); print_eos(lat, opt_form); }
 /*
  * print_best_path()
  */
 static void
-print_best_path(int opt_form, char *format)
+print_best_path(cha_lat_t *lat, int opt_form, char *format)
 {
     int i, last, pbuf_last, isfirst = 1;
     int path_num_composit = 0;
@@ -1158,9 +1164,10 @@ print_best_path(int opt_form, char *format)
     char pron[CHA_INPUT_SIZE];
     char base[CHA_INPUT_SIZE];
     mrph_data_t composit_mrph;
+    mrph_t m;
     mrph_t *cur_mrph, *pre_mrph;
 
-    print_bos(opt_form);
+    print_bos(lat, opt_form);
 
     last = Cha_path[Cha_path_num - 1].best_path;
 
@@ -1177,7 +1184,8 @@ print_best_path(int opt_form, char *format)
      */
     /* initialization */
 
-    composit_mrph.posid = 0;
+    composit_mrph.mrph = &m;
+    composit_mrph.mrph->posid = 0;
     composit_mrph.reading = reading;
     composit_mrph.pron = pron;
     composit_mrph.base = base;
@@ -1199,7 +1207,7 @@ print_best_path(int opt_form, char *format)
 	    && Cha_hinsi[cur_mrph->posid].composit
 	    && (Cha_hinsi[cur_mrph->posid].composit == Cha_hinsi[pre_mrph->posid].composit)) {
 
-	    if (!composit_mrph.posid) 
+	    if (!composit_mrph.mrph->posid) 
 	        path_num_composit = path_buffer[i];
 
 	    concat_composit_mrph(&composit_mrph, &data);
@@ -1211,14 +1219,15 @@ print_best_path(int opt_form, char *format)
 		else
 		    cha_putc(',', cha_output);
 	    }
-	    if (!composit_mrph.posid)
-		print_mrph(path_buffer[i], &data, format);
+	    if (!composit_mrph.mrph->posid)
+		print_mrph(lat, path_buffer[i], &data, format);
 	    else {
                 concat_composit_mrph_end(&composit_mrph, &data);
                 Cha_path[path_num_composit].end =
-                   Cha_path[path_num_composit].start + composit_mrph.headword_len;
-		print_mrph(path_num_composit, &composit_mrph, format);
-		composit_mrph.posid = 0;
+                   Cha_path[path_num_composit].start 
+		    + composit_mrph.mrph->headword_len;
+		print_mrph(lat, path_num_composit, &composit_mrph, format);
+		composit_mrph.mrph->posid = 0;
 	    }
 	}
 	cur_mrph = pre_mrph;
@@ -1251,7 +1260,7 @@ collect_all_mrph(int path_num)
 }
 
 static void
-print_all_mrph(int opt_form, char *format)
+print_all_mrph(cha_lat_t *lat, int opt_form, char *format)
 {
     int i;
     int isfirst = 1;		/* ʸƬɤΥե饰 for -d option */
@@ -1268,7 +1277,7 @@ print_all_mrph(int opt_form, char *format)
 	Cha_path[Cha_path_num - 1].do_print = 2;
     }
 
-    print_bos(opt_form);
+    print_bos(lat, opt_form);
     for (i = 0; i < Cha_path_num; i++) {
 	if (Cha_path[i].do_print) {
 	    if (opt_form == 'd') {
@@ -1277,18 +1286,18 @@ print_all_mrph(int opt_form, char *format)
 		else
 		    cha_putc(',', cha_output);
 	    }
-	    print_path_mrph(i, format);
+	    print_path_mrph(lat, i, format);
 	}
     }
-    print_anno(Cha_path_num - 1, format);
-    print_eos(opt_form);
+    print_anno(lat, Cha_path_num - 1, format);
+    print_eos(lat, opt_form);
 }
 
 /*
  * print_all_path()
  */
 static void
-print_all_path_sub(int path_num, int paths, char *format)
+print_all_path_sub(cha_lat_t *lat, int path_num, int paths, char *format)
 {
     int i, j;
 
@@ -1296,40 +1305,40 @@ print_all_path_sub(int path_num, int paths, char *format)
 	if (Cha_path[path_num].path[0] == 0) {
 	    pos_end = 0;
 	    for (j = paths - 1; j >= 0; j--)
-		print_path_mrph(path_buffer[j], format);
-	    print_anno(Cha_path_num - 1, format);
+		print_path_mrph(lat, path_buffer[j], format);
+	    print_anno(lat, Cha_path_num - 1, format);
 	    cha_puts("EOP\n", cha_output);
 	} else {
 	    path_buffer[paths] = Cha_path[path_num].path[i];
-	    print_all_path_sub(Cha_path[path_num].path[i], paths + 1,
+	    print_all_path_sub(lat, Cha_path[path_num].path[i], paths + 1,
 			       format);
 	}
     }
 }
 
 static void
-print_all_path(int opt_form, char *format)
+print_all_path(cha_lat_t *lat, int opt_form, char *format)
 {
-    print_bos(opt_form);
-    print_all_path_sub(Cha_path_num - 1, 0, format);
-    print_eos(opt_form);
+    print_bos(lat, opt_form);
+    print_all_path_sub(lat, Cha_path_num - 1, 0, format);
+    print_eos(lat, opt_form);
 }
 
 void
-cha_print_path(int opt_show, int opt_form, char *format)
+cha_print_path(cha_lat_t *lat, int opt_show, int opt_form, char *format)
 {
     if (opt_form == 'd')
 	cha_putc('[', cha_output);
 
     switch (opt_show) {
     case 'm':
-	print_all_mrph(opt_form, format);
+	print_all_mrph(lat, opt_form, format);
 	break;
     case 'p':
-	print_all_path(opt_form, format);
+	print_all_path(lat, opt_form, format);
 	break;
     default:
-	print_best_path(opt_form, format);	/* 'b' */
+	print_best_path(lat, opt_form, format);	/* 'b' */
     }
 
     if (opt_form == 'd')
@@ -1340,8 +1349,8 @@ void
 cha_print_bos_eos(int opt_form)
 {
     pos_end = 0;
-    print_bos(opt_form);
-    print_eos(opt_form);
+    print_bos(NULL, opt_form);
+    print_eos(NULL, opt_form);
 }
 
 void
