@@ -1,5 +1,5 @@
 #
-# $Id: ChaSen.pm,v 1.1.1.1 2007/03/13 07:40:10 masayu-a Exp $
+# $Id: ChaSen.pm,v 1.2 2008/07/22 13:00:45 masayu-a Exp $
 
 package Text::ChaSen;
 
@@ -23,13 +23,11 @@ Text::ChaSen - ChaSen library module for perl
 
 =item getopt_argv($arg1, $arg2, ...)
 
-䦤˥ץϤԤץϡchasenޥ
+䥤˥ץϤԤץϡchasenޥ
 ˻ǤΤ˽ऺ뤬C<-s>C<-D>ʤɤΥФ䥯饤Ȥ˴ؤ
 ץѤǤʤ
 ޤֺǽΥץϥץΥե̾Ǥ롣
 
-=back
-
 㤨СΤ褦ʰǼ¹Ԥchasenޥɤ롣
 
 $ chasen C<-j> C<-F> '%m '
@@ -47,6 +45,8 @@ getopt_argv('chasen', 'C<-j>', 'C<-F>', '%m ');
 ǲϤԤ̤ʸȤ֤δؿϡΥС
 ȤθߴΤ¸ߤ롣
 
+=back
+
 =head1 COPYRIGHT
 
 Copyright(c) 1998, 1999 NOKUBI Takatsugu <knok@daionet.gr.jp>
@@ -132,7 +132,7 @@ require DynaLoader;
 %EXPORT_TAGS = (all => [qw(getopt_argv sparse_tostr fparse_tostr
  sparse_tostr_long)]);
 
-$VERSION = '1.03';
+$VERSION = '1.04';
 
 bootstrap Text::ChaSen $VERSION;
 
